/*
uNavINS_MPU9250.ino
Brian R Taylor
brian.taylor@bolderflight.com

Copyright (c) 2017 Bolder Flight Systems

Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
and associated documentation files (the "Software"), to deal in the Software without restriction, 
including without limitation the rights to use, copy, modify, merge, publish, distribute, 
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or 
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "uNavINS.h"
#include "MPU9250.h"
#include "UBLOX.h"
#include "EEPROM.h"
#include "A_configDefines.h"
#include "Eigen.h"
#include <Eigen/Dense>

// an MPU-9250 object on SPI bus 0 with chip select 24
//MPU9250 Imu(SPI,24);
#if defined(TEENSYDUINO)
  MPU9250 Imu(IMU_BUS, IMU_ADDR, IMU_SCL, IMU_SDA, IMU_SPD);
#else
  MPU9250 Imu(IMU_BUS, IMU_ADDR);
#endif
int status;

// a flag for when the MPU-9250 has new data
volatile int newIMUData;
// a uNavINS object
uNavINS Filter;
UBLOX gps(GPS_Start);
bool newGpsData;
unsigned long prevTOW, currentTOW;
gpsData uBloxData;

// EEPROM buffer and variables to load accel and mag bias 
// and scale factors from CalibrateMPU9250.ino
uint8_t EepromBuffer[48];
float axb, axs, ayb, ays, azb, azs;
float hxb, hxs, hyb, hys, hzb, hzs;

// timers to measure performance
unsigned long nextPrintTimeMillis, currentTimeMillis;
float rtcOffset, rtcTime, newIMUDataTimer;

// D2R and R2D
const float D2R = PI / 180.0f;
const float R2D = 180.0f / PI;
    
void setup() {
  // serial to display data
  Serial.begin(115200);
  Serial.println(" ");
  Serial.println("Serial Initialized");
  while(!Serial) {}

  gps.begin(GPS_BAUD);
  
  // start communication with IMU 
  status = Imu.begin();
  if (status < 0) {
    Serial.println("IMU initialization unsuccessful");
    Serial.println("Check IMU wiring or try cycling power");
    Serial.print("Status: ");
    Serial.println(status);
    while(1) {}
  }

  //Filter.setInitializationDuration(120000000);
  // setting a 41 Hz DLPF bandwidth
  //Imu.setDlpfBandwidth(MPU9250::DLPF_BANDWIDTH_20HZ);
  Imu.setDlpfBandwidth(MPU9250::DLPF_BANDWIDTH_41HZ);
  // setting SRD to 9 for a 100 Hz update rate
  Imu.setSrd(IMU_SRD);

  // Load in calibration biases
  loadCalValues();
  
  // enabling the data ready interrupt
  Imu.enableDataReadyInterrupt();
  // attaching the interrupt to microcontroller pin 1
  pinMode(IMU_INT_PIN,INPUT);
  attachInterrupt(IMU_INT_PIN,runFilter,RISING);

  // Initialize rtcOffset
  rtcOffset = (float) micros() / 1000000.0;
  // Initialize timing variables for printing
  currentTimeMillis = millis();
  nextPrintTimeMillis = millis();

  Serial.println("Setup Complete");
}

void loop() {

  //Define RTC time for this iteration
  rtcTime = (float) micros() / 1000000.0 - rtcOffset;
    
  gps.read(&uBloxData);
  //Serial.println("Read GPS");
  //Serial.println(uBloxData.numSV);

  // If GPS receiver is locked to at least five sats and if there is new
  // IMU data, then run the EKF. This seems to run at about 50Hz (20ms samples).
  if (uBloxData.numSV > 5) {
    //Serial.println("numSV>5");

    // Print out warning if apparent loss of IMU data
    if ((rtcTime - newIMUDataTimer) > 0.150){
      Serial.println(" WARNING: Apparent loss of IMU data !!");
    }
    
    if (newIMUData == 1) {
      //Serial.println("newData=1");
      newIMUData = 0;
      newIMUDataTimer = (float) micros() / 1000000.0f - rtcOffset;

      // read the IMU sensor
      Imu.readSensor();
      //Serial.println("IMU sensor read");
            
      Filter.update(uBloxData.iTOW,uBloxData.velN,uBloxData.velE,uBloxData.velD,
                    uBloxData.lat*D2R,uBloxData.lon*D2R,uBloxData.hMSL,
                    Imu.getGyroX_rads(),Imu.getGyroY_rads(),Imu.getGyroZ_rads(),
                    Imu.getAccelX_mss(),Imu.getAccelY_mss(),Imu.getAccelZ_mss(),
                    Imu.getMagX_uT(),Imu.getMagY_uT(),Imu.getMagZ_uT());
      
      
      // Print values to serial
      serialPrint();

    }
  }
}

void runFilter() {
  newIMUData = 1;
}
